
from enum import Enum

from django.test import TestCase
from django.db import models
from django.db.utils import DataError
from django_fake_model import models as fake_models
from django.core.exceptions import ValidationError
import pytest

from lily.base.models import (
    ImmutableModel,
    JSONSchemaField,
    ValidatingModel,
    EnumChoiceField,
    enum,
    array,
    object,
    string,
)


class SchemaHelpersTestCase(TestCase):

    #
    # ENUM
    #
    def test_enum__from_list(self):

        assert enum('Jack', 'Alice') == {
            'enum': ['Jack', 'Alice'],
            'enum_name': None,
            'type': 'string',
        }
        assert enum(1, 11) == {
            'enum': [1, 11],
            'enum_name': None,
            'type': 'integer',
        }

    def test_enum__from_enum(self):

        class Names(Enum):
            JACK = 'Jack'

            ALICE = 'Alice'

        assert enum(Names) == {
            'enum': ['Jack', 'Alice'],
            'enum_name': 'Names',
            'type': 'string',
        }

        class Ages(Enum):
            kid = 10

            adult = 21

        assert enum(Ages) == {
            'enum': [10, 21],
            'enum_name': 'Ages',
            'type': 'integer',
        }


class ImmutableEntity(fake_models.FakeModel, ImmutableModel):

    name = models.CharField(max_length=10)


@ImmutableEntity.fake_me
class ImmutableModelTestCase(TestCase):

    def test_save_not_allowed_for_existing_instance(self):

        e = ImmutableEntity(name='john')
        e.save()

        with pytest.raises(ImmutableModel.ModelIsImmutableError):
            e.name = 'mark'
            e.save()

    def test_create_allowed(self):

        e = ImmutableEntity.objects.create(name='john')

        assert e.name == 'john'

    def test_validates_on_save(self):

        with pytest.raises(ValidationError) as e:
            ImmutableEntity(name=11 * 'a').save()

        assert e.value.message_dict == {
            'name': [
                'Ensure this value has at most 10 characters (it has 11).',
            ],
        }


class JSONModel(fake_models.FakeModel):

    SCHEMA = {
        '$schema': 'http://json-schema.org/draft-04/schema#',
        'type': 'array',
        'minItems': 1,
        'items': {
            'type': 'object',
            'properties': {
                'answer': {
                    'type': 'string',
                },
                'correct': {
                    'type': 'boolean',
                },
            },
            'required': ['answer', 'correct'],
        },
    }

    answers = JSONSchemaField(schema=SCHEMA)


@JSONModel.fake_me
class JSONSchemaFieldTestCase(TestCase):

    def test_valid_model(self):

        m = JSONModel(answers=[
            {'answer': 'hello', 'correct': True},
        ])
        m.save()

        assert m.answers == [
            {'answer': 'hello', 'correct': True},
        ]

    def test_invalid_model__empty_array(self):

        with pytest.raises(ValidationError) as e:
            JSONModel(answers=[]).clean_fields()

        error = e.value.error_dict['answers'][0]
        assert error.message == 'This field cannot be blank.'

    def test_invalid_model__not_array(self):

        with pytest.raises(ValidationError) as e:
            JSONModel(answers={'hi': 'there'}).clean_fields()

        error = e.value.error_dict['answers'][0]
        assert error.message == (
            "JSON did not validate. PATH: '.' REASON: {'hi': 'there'} "
            "is not of type 'array'"
        )

    def test_invalid_model__invalid_items__wrong_fields(self):

        with pytest.raises(ValidationError) as e:
            JSONModel(answers=[
                {'answer': 'hi', 'correct': True},
                {'NOT_answer': 'hi', 'WHAT_correct': True},
            ]).clean_fields()

        error = e.value.error_dict['answers'][0]
        assert error.message == (
            "JSON did not validate. "
            "PATH: '1' REASON: 'answer' is a required property"
        )


class EnumModel(fake_models.FakeModel):

    answer = EnumChoiceField(
        enum_name='super_enum',
        max_length=8,
        choices=[('A', 'A'), ('B', 'B')])

    class Names(Enum):

        ALICE = 'ALICE'

        JACK = 'JACK'

    name = EnumChoiceField(enum=Names, max_length=8)


@EnumModel.fake_me
class EnumChoiceFieldTestCase(TestCase):

    def test_model(self):

        m = EnumModel(answer='A', name='JACK')
        m.save()

        assert m.answer == 'A'
        assert m.name == 'JACK'

    def test_extra_field_attributes(self):

        for field in EnumModel._meta.fields:
            if field.name == 'answer':
                assert field.enum_name == 'super_enum'

            if field.name == 'name':
                assert field.enum_name == 'Names'
                assert field.choices == [('ALICE', 'ALICE'), ('JACK', 'JACK')]


class ValidatingEntity(fake_models.FakeModel, ValidatingModel):

    name = models.CharField(max_length=10)


class ValidatingEntityWithJSON(fake_models.FakeModel, ValidatingModel):

    tasks = JSONSchemaField(
        schema=array(
            object(
                name=string())))

    def clean(self):
        if len(set([t['name'] for t in self.tasks])) < 2:
            raise ValidationError('at least 2 unique task needed')


@ValidatingEntity.fake_me
@ValidatingEntityWithJSON.fake_me
class ValidatingModelTestCase(TestCase):

    #
    # ValidatingEntity
    #
    def test_validates_on_save(self):

        with pytest.raises(ValidationError) as e:
            ValidatingEntity(name=11 * 'a').save()

        assert e.value.message_dict == {
            'name': [
                'Ensure this value has at most 10 characters (it has 11).',
            ],
        }

    def test_validates_on_create(self):

        with pytest.raises(ValidationError) as e:
            ValidatingEntity.objects.create(name=11 * 'a')

        assert e.value.message_dict == {
            'name': [
                'Ensure this value has at most 10 characters (it has 11).',
            ],
        }

    def test_validates_on_bulk_create(self):

        with pytest.raises(DataError):
            ValidatingEntity.objects.bulk_create([
                ValidatingEntity(name=11 * 'a'),
            ])

    #
    # ValidatingEntityWithJSON
    #
    def test_schema_validation__run_before_clean(self):

        with pytest.raises(ValidationError) as e:
            ValidatingEntityWithJSON.objects.create(tasks={'what': 'hey'})

        assert e.value.message_dict == {
            'tasks': [
                "JSON did not validate. PATH: '.' REASON: {'what': 'hey'} "
                "is not of type 'array'",
            ],
        }

    def test_schema_validation_when_ok_clean_is_called(self):

        with pytest.raises(ValidationError) as e:
            ValidatingEntityWithJSON.objects.create(tasks=[{'name': 'hey'}])

        assert e.value.message_dict == {
            '__all__': ['at least 2 unique task needed'],
        }




from django.contrib.postgres.search import (
    SearchVector,
    SearchVectorCombinable,
    CombinedExpression,
)
from django.db import connection
from trans import trans

from .latex.transformer import transform as transform_latex
from .constants import HASHTAG_ESCAPE_SEQUENCE, HASHTAG_PATTERN


def to_tsvector(conf, text, weight=None):
    if text is None:
        return text

    def replace_hashtag(match):

        return "{prefix}{escape_seq}{text}".format(
            prefix=match.group('prefix'),
            escape_seq=HASHTAG_ESCAPE_SEQUENCE,
            text=match.group('text'))

    text = HASHTAG_PATTERN.sub(replace_hashtag, text)
    text = transform_latex(text, conf)

    with connection.cursor() as c:
        # deal with accents
        c.execute(
            "SELECT token FROM ts_debug(%s, %s) WHERE alias != 'blank'",
            [conf, text])
        tokens = [t[0] for t in c.fetchall()]

        escaped_tsvector_part = ''
        if tokens:
            for i, token in enumerate(tokens):
                escaped_token = trans(token.lower())
                if escaped_token != token:
                    escaped_tsvector_part += " '{token}':{position}".format(
                        token=escaped_token,
                        position=i + 1)

        if not weight:
            c.execute("SELECT to_tsvector(%s, %s)", [conf, text])

        else:
            c.execute(
                "SELECT setweight(to_tsvector(%s, %s), %s)",
                [conf, text, weight])

        tsvector = c.fetchone()[0]

        tsvector += escaped_tsvector_part
        return tsvector


def _concatenate_tsvectors(v1, v2):

    with connection.cursor() as c:
        c.execute("SELECT %s::tsvector || %s::tsvector", [v1, v2])

        return c.fetchone()[0]


def concatenate_tsvectors(*vectors):

    v1 = vectors[0]
    for v in vectors[1:]:
        v1 = _concatenate_tsvectors(v1, v)

    return v1


class VectorCombinable(SearchVectorCombinable):

    def _combine(self, other, connector, reversed, node=None):
        return CombinedVector(self, connector, other, self.config)


class CombinedVector(VectorCombinable, CombinedExpression):
    def __init__(self, lhs, connector, rhs, config, output_field=None):
        self.config = config
        super(
            CombinedVector,
            self).__init__(lhs, connector, rhs, output_field)


class StoredVector(SearchVector, VectorCombinable):
    """Represent stored text-search vector.

    Used for wrapping either stored ``tsvector`` column or the aggregate of
    such columns.

    """

    def as_sql(self, compiler, connection, function=None, template=None):

        return super(StoredVector, self).as_sql(
            compiler,
            connection,
            function=None,
            template='%(expressions)s')


class OnTheFlyVector(SearchVector, VectorCombinable):
    """Represent on-the-fly calculated text-search vector.

    Used for parsing supplied column in on the fly fashion to ``tsvector``
    and optionally to combine itself with other tsvectors (either on the fly
    ones or stored ones).

    """




from unittest.mock import Mock, call

import pytest

import search


def test_stored_vector__to_sql(mocker):

    compiler, connection = Mock(), Mock()
    as_sql = mocker.patch('search.vector.SearchVector.as_sql')

    vector = search.StoredVector('column')
    result = vector.as_sql(compiler, connection)

    assert (
        as_sql.call_args_list ==
        [call(
            compiler,
            connection,
            function=None,
            template='%(expressions)s')])
    assert result == as_sql.return_value


@pytest.mark.parametrize("conf, text, expected", [

    # case 0 - none input, none output
    ('simple', None, None),

    # case 1 - empty string input, empty string output
    ('simple', '', ''),

    # case 2 - single word
    ('simple', 'hello', "'hello':1"),

    # case 3 - many words case
    (
        'simple',
        'hi you how are you?',
        "'hi':1 'you':2,5 'how':3 'are':4",
    ),

    # case 4 - many words case with single hashtag
    (
        'simple',
        'hi #enjoyTheSilence how are you?',
        "'hi':1 '007enjoythesilence':2 'how':3 'are':4 'you':5",
    ),

    # case 5 - many hashtags
    (
        'simple',
        'hi #enjoyTheSilence how are #YouWin #watIsThis?',
        "'hi':1 '007enjoythesilence':2 'how':3 'are':4 '007youwin':5 "
        "'007watisthis':6",
    ),

    # case 6 - transform latex
    (
        'simple',
        r'#enjoyTheSilence $$ \frac{a}{b^2} $$',
        "'007enjoythesilence':1 'fraction':2 'a':3 'divided':4 'by':5 'b':6 "
        "'squared':7",
    ),

    #
    # Language Specific Cases - POLISH
    #
    # case 7 - polish - escape accents
    (
        'polish',
        'część grzegżółek dość Świata',
        "'część':1 'czesc':1 'grzegżółek':2 'grzegzolek':2 'dość':3 "
        "'dosc':3 'świata':4 'swiata':4",
    ),

    # case 8 - polish - escape accents - triangulation
    pytest.param(
        'polish',
        'Pchnąć w tę łódź jeża lub ośm skrzyń fig',
        "'pchnąć':1 'pchnac':1 'w':2 'te':3 'tę':3 'łódź':4 'lodz':4 'jeża':5 "
        "'jeż':5 'jeza':5 'lub':6 'luba':6 'lubić':6 'ośm':7 'osm':7 "
        "'skrzynia':8 'skrzyn':8 'figa':9 'figi':9",
        marks=pytest.mark.xfail(
            reason='it will work again after hunspell integration')
    ),

    # case 9 - many hashtags - polish
    pytest.param(
        'polish',
        'część grzegżółek #dośćŚwiata #wokółNAS słyszałeś',
        "'część':1 'czesc':1 'grzegżółek':2 'grzegzolek':2 '007dośćświata':3 "
        "'007doscswiata':3 '007wokółnas':4 '007wokolnas':4 'slyszales':5 "
        "'słyszeć':5",
        marks=pytest.mark.xfail(
            reason='it will work again after hunspell integration')
    ),

    # case 10 - latex transformation - poi
    pytest.param(
        'polish',
        'rowiązanie to $$ \\sqrt{a ^ 2} $$',
        "'rowiązanie':1 'rowiazanie':1 'ten':2 'to':2 'sqrt':3 'funkcja':4 "
        "'pierwiastek':5 'pierwiastka':5 'kwadratowy':6 'z':7 'a':8 'do':9 "
        "'kwadrat':10",
        marks=pytest.mark.xfail(
            reason='it will work again after hunspell integration')
    ),

], ids=list(map(str, range(0, 11))))
@pytest.mark.django_db
def test_to_tsvector(conf, text, expected, mocker):

    mocker.patch('search.vector.HASHTAG_ESCAPE_SEQUENCE', '007')

    def transform(x):
        if not x:
            return x

        return set(x.split(' '))

    assert transform(search.to_tsvector(conf, text)) == transform(expected)


@pytest.mark.parametrize("v1, v2, v3, expected", [

    # case 0 - single word vectors
    ("'cat':1", "'dog':1", None, "'cat':1 'dog':2"),

    # case 1 - single word vectors with weights
    ("'cat':1B", "'dog':1C", None, "'cat':1B 'dog':2C"),

    # case 2 - many words
    (
        "'cat':1 'whatever':11",
        "'dog':9 'hospital':11 'life':13",
        None,
        "'cat':1 'whatever':11 'dog':20 'hospital':22 'life':24",
    ),

    # case 3 - many words with weights
    (
        "'cat':1B 'whatever':11C",
        "'dog':9 'hospital':11A 'life':13",
        None,
        "'cat':1B 'whatever':11C 'dog':20 'hospital':22A 'life':24",
    ),

    # case 4 - three vectors
    (
        "'cat':1B 'whatever':11C",
        "'dog':9 'hospital':11A 'life':13",
        "'whatever':1B",
        "'cat':1B 'whatever':11C,25B 'dog':20 'hospital':22A 'life':24",
    ),

], ids=list(map(str, range(0, 5))))
@pytest.mark.django_db
def test_concatenate(v1, v2, v3, expected):

    def transform(x):
        if not x:
            return x

        return set(x.split(' '))

    if v3:
        vectors = [v1, v2, v3]

    else:
        vectors = [v1, v2]

    assert (
        transform(search.concatenate_tsvectors(*vectors)) ==
        transform(expected))




from rest_framework import serializers as drf_serializers

# -- needed for correct namespacing so that one could access all
# -- serializers stuff as a part of the `parsers` namespace
from .serializers import *  # noqa


class QueryParser(drf_serializers.Serializer):

    def __init__(self, *args, **kwargs):

        raw_data = kwargs.get('data')

        if raw_data is not None:
            data = {}
            for field_name, field_class in self.fields.items():

                if not isinstance(field_class, ListField):  # noqa
                    try:
                        data[field_name] = raw_data[field_name]

                    # -- ignore key errors since they all will be caught
                    # -- by the validation
                    except (KeyError, IndexError):
                        pass

                else:
                    try:
                        data[field_name] = raw_data.getlist(field_name)

                    except AttributeError:
                        data[field_name] = raw_data.get(field_name)

            kwargs['data'] = data
            super(QueryParser, self).__init__(*args, **kwargs)


class PageQueryParser(QueryParser):

    offset = IntegerField(default=0)  # noqa

    limit = IntegerField(default=100)  # noqa


class FullTextSearchQueryParser(QueryParser):

    query = CharField(default=None)  # noqa


class BodyParser(drf_serializers.Serializer):
    pass


class ModelParser(drf_serializers.ModelSerializer):

    serializer_choice_field = EnumChoiceField  # noqa

    def build_standard_field(self, field_name, model_field):

        from . import models  # noqa - avoid circular dependency

        field_class, field_kwargs = super(
            ModelParser, self).build_standard_field(
                field_name, model_field)

        if isinstance(model_field, models.EnumChoiceField):
            field_kwargs['enum_name'] = model_field.enum_name
            field_class = EnumChoiceField  # noqa

        return field_class, field_kwargs
